// SPDX-FileCopyrightText: 2018 Paul Schaub <vanitasvitae@fsfe.org>
//
// SPDX-License-Identifier: Apache-2.0

/**
 * PGPainless - Use OpenPGP Painlessly!
 *
 * @see <a href="http://pgpainless.org">org.pgpainless.core.org</a>
 */
package org.pgpainless;
