// SPDX-FileCopyrightText: 2021 Paul Schaub <vanitasvitae@fsfe.org>
//
// SPDX-License-Identifier: Apache-2.0

/**
 * PGPainless SOP implementing a Stateless OpenPGP Command Line Interface.
 * @see <a href="https://tools.ietf.org/html/draft-dkg-openpgp-stateless-cli-01">
 *     Stateless OpenPGP Command Line Interface
 *     draft-dkg-openpgp-stateless-cli-01</a>
 */
package org.pgpainless.cli;
