// SPDX-FileCopyrightText: 2020 Paul Schaub <vanitasvitae@fsfe.org>
//
// SPDX-License-Identifier: Apache-2.0

/**
 * Classes that allow setting a custom implementation of {@link java.security.Provider}.
 */
package org.pgpainless.provider;
