// SPDX-FileCopyrightText: 2022 Paul Schaub <vanitasvitae@fsfe.org>
//
// SPDX-License-Identifier: Apache-2.0

/**
 * Pushdown Automaton to verify validity of packet sequences according to the OpenPGP Message format.
 */
package org.pgpainless.decryption_verification.syntax_check;
