// SPDX-FileCopyrightText: 2018 Paul Schaub <vanitasvitae@fsfe.org>
//
// SPDX-License-Identifier: Apache-2.0

/**
 * Different Key Ring Selection Strategies.
 */
package org.pgpainless.util.selection.keyring;
