// SPDX-FileCopyrightText: 2020 Paul Schaub <vanitasvitae@fsfe.org>
//
// SPDX-License-Identifier: Apache-2.0

/**
 * Passphrase Provider classes.
 */
package org.pgpainless.key.protection.passphrase_provider;
