// SPDX-FileCopyrightText: 2022 Paul Schaub <vanitasvitae@fsfe.org>
//
// SPDX-License-Identifier: Apache-2.0

/**
 * Utility classes related to creating keys with GNU DUMMY S2K values.
 */
package org.gnupg;
