<!doctype html>
<html>
  <head>
    <meta charset="utf-8">
    <meta http-equiv="X-UA-Compatible" content="chrome=1">
    <title>{{ site.title | default: site.github.repository_name }} by {{ 
site.github.owner_name }}</title>

    <link rel="stylesheet" href="{{ '/assets/css/style.css?v=' | append: 
site.github.build_revision | relative_url }}">
    <meta name="viewport" content="width=device-width">
    <!--[if lt IE 9]>
    <script 
src="//html5shiv.googlecode.com/svn/trunk/html5.js"></script>
    <![endif]-->
  </head>
  <body>
    <div class="wrapper" style="width: 1060px">
      <header>
	{% if site.logo %}
          <img src="{{site.logo | relative_url}}" alt="Logo" />
	{% endif %}
        <p>{{ site.description | default: site.github.project_tagline 
}}</p>

        <a href="https://pgpainless.org">Home</a>
        <br>
	    <a href="https://search.maven.org/search?q=g:org.pgpainless%20AND%20a:pgpainless-core&core=gav">Releases</a>
	    <br>
        <a href="https://pgpainless.rtfd.io">Documentation</a>
        <br>
        <a href=" https://javadoc.io/doc/org.pgpainless ">Javadoc</a>
        <br>
        <a href="https://coveralls.io/github/pgpainless/pgpainless">Coverage</a>
        <br>


        {% if site.github.is_project_page %}
          <p class="view"><a href="{{ site.github.repository_url 
}}">View the Project on GitHub <small>{{ github_name }}</small></a></p>
        {% endif %}

        {% if site.github.is_user_page %}
          <p class="view"><a href="{{ site.github.owner_url }}">View My 
GitHub Profile</a></p>
        {% endif %}

        {% if site.show_downloads %}
          <ul>
            <li><a href="{{ site.github.zip_url }}">Download <strong>ZIP 
File</strong></a></li>
            <li><a href="{{ site.github.tar_url }}">Download <strong>TAR 
Ball</strong></a></li>
            <li><a href="{{ site.github.repository_url }}">View On 
<strong>GitHub</strong></a></li>
          </ul>
        {% endif %}
      </header>
      <section style="width: 780px">

      {{ content }}

      </section>
      <footer>
        {% if site.github.is_project_page %}
        <p>This project is maintained by <a href="{{ 
site.github.owner_url }}">{{ site.github.owner_name }}</a></p>
        {% endif %}
        <p><small>Hosted on GitHub Pages &mdash; Theme by <a 
href="https://github.com/orderedlist">orderedlist</a></small></p>
      </footer>
    </div>
    <script src="{{ '/assets/js/scale.fix.js' | relative_url 
}}"></script>
  </body>
</html>
