// SPDX-FileCopyrightText: 2021 Paul Schaub <vanitasvitae@fsfe.org>
//
// SPDX-License-Identifier: Apache-2.0

package org.pgpainless.exception;

import java.io.IOException;

/**
 * Exception that gets thrown when the verification of a modification detection code failed.
 */
public class ModificationDetectionException extends IOException {

}
