// SPDX-FileCopyrightText: 2023 Paul Schaub <vanitasvitae@fsfe.org>
//
// SPDX-License-Identifier: Apache-2.0

package sop.testsuite.pgpainless.operation;

import sop.testsuite.operation.InlineSignInlineVerifyTest;

public class PGPainlessInlineSignInlineVerifyTest extends InlineSignInlineVerifyTest {

}
